
import sys
from typing import List, Tuple
from pathlib import Path

import git
import yaml


def init(repo_path: Path) -> None:
    """
    Initialize coldbrew (create initial coldbrew branch).
    Preconditions: 1. At least one commit in the repo and 2. No coldbrew branches yet
    :param repo_path: Path to directory containing a .git/ directory
    :return:
    """

    repo = git.Repo(repo_path)

    if len(repo.heads) == 0:
        print("Repository does not contain any commits. Make an initial commit to create a head which coldbrew can switch back to")
        print("Aborted.")
        sys.exit(1)

    if any(['coldbrew/' in head.name for head in repo.heads]):
        print("coldbrew has already been initialized for this repository")
        print("Aborted.")
        sys.exit(1)

    repo.create_head('coldbrew/util')


def create(repo_path: Path, description: str, group: str) -> None:
    """
    Creates a new experiment by creating an orphaned commit on the coldbrew/util branch.
    Then, a new branch is created, called coldbrew/<hexsha of that commit>.
    That branch points to the newly created commit.
    :param repo_path: Path to the git repository's working directory
    :param description: Description of the experiment
    :param group: Experiment's group
    :return:
    """
    repo = git.Repo(repo_path)

    # Store current state: Untracked files and HEAD
    untracked_files = repo.untracked_files
    current_head = repo.head.reference

    # Create commit on coldbrew util branch
    repo.head.reference = repo.heads['coldbrew/util']
    repo.git.add(all=True)
    message = yaml.dump({'description': description, 'group': group})
    coldbrew_commit = repo.index.commit(message, parent_commits=[])

    # Create new branch for that commit
    repo.create_head(f'coldbrew/{coldbrew_commit.hexsha}', commit=coldbrew_commit)

    # Unstage previously untracked files and reset head
    repo.head.reference = current_head
    repo.head.reset(paths=untracked_files)

    # Print infos
    print(f"\N{check mark} Created experiment '{description}' (group '{group}') with ID {coldbrew_commit.hexsha}")


def iterate(repo_path: Path, experiment_id: str, description: str) -> None:
    """
    Iterates on an existing experiment by creating a new commit on the parent commit's branch
    :param repo_path: Path to the git repository's working directory
    :param experiment_id: The ID of the experiment's commit (or the beginning thereof, as long as it is unique)
    :param description: Description of the iteration, e.g. diff
    :return:
    """
    repo = git.Repo(repo_path)

    # Check if ID exists and is unique
    selected_experiments = [head for head in repo.heads if f'coldbrew/{experiment_id}' in head.name]
    if len(selected_experiments) == 0:
        print(f"Experiment with ID {experiment_id} does not exist")
        print("Aborted.")
        sys.exit(1)
    elif len(selected_experiments) > 1:
        print(f"More than one experiment found for ID {experiment_id}. Specify a unique identifier")
        print("Aborted.")
        sys.exit(1)

    experiment_head = selected_experiments[0]

    # Store current state: Untracked files and HEAD
    untracked_files = repo.untracked_files
    current_head = repo.head.reference

    # Create commit on experiment's coldbrew branch
    repo.head.reference = experiment_head
    repo.git.add(all=True)
    message = yaml.dump({'description': description})
    coldbrew_commit = repo.index.commit(message)

    # Unstage previously untracked files and reset head
    repo.head.reference = current_head
    repo.head.reset(paths=untracked_files)

    # Print infos
    print(f"\N{check mark} Created iteration '{description}' ({coldbrew_commit.hexsha}) on experiment {experiment_head.name.split('/')[1]}")


def push(repo_path: Path) -> None:
    """
    Pushes all coldbrew branches to all coldbrew remotes
    :param repo_path: Path to the git repository's working directory
    :return:
    """
    repo = git.Repo(repo_path)

    # Push coldbrew branches
    coldbrew_remotes = [remote for remote in repo.remotes if 'coldbrew-' in remote.name]
    for remote in coldbrew_remotes:
        remote.push(refspec=f"refs/heads/coldbrew/*:refs/heads/coldbrew/*")

    print(f"\N{check mark} Pushed to {', '.join([next(remote.urls) for remote in coldbrew_remotes])}")


def contains_git_repo(repo_path: Path) -> bool:
    """
    Check whether given directory contains a valid git repository
    :param repo_path: Path to directory
    :return:
    """
    try:
        git.Repo(repo_path)
    except git.InvalidGitRepositoryError:
        return False
    return True


def remote_reachable(repo_path: Path, name: str) -> bool:
    """
    Check whether the given remote is reachable
    :param repo_path: Path to directory containing a .git/ directory
    :return:
    """
    repo = git.Repo(repo_path)

    try:
        repo.git.ls_remote(name)
        return True
    except git.GitCommandError:
        return False


def remote_add(repo_path: Path, name: str, url: str) -> None:
    """
    Adds a new git remote, prefixed by 'coldbrew-'
    :param repo_path: Path to directory containing a .git/ directory
    :param name: Name of the remote
    :param url: URL of the remote
    :return:
    """
    repo = git.Repo(repo_path)
    repo.create_remote('coldbrew-'+name, url)
    if not remote_reachable(repo_path, 'coldbrew-'+name):
        print(f"Remote {name} is not reachable. Use 'coldbrew remote add' to try again with a new url")
        repo.delete_remote('coldbrew-'+name)
        sys.exit(1)


def remote_remove(repo_path: Path, name: str) -> None:
    """
    Remove an existing coldbrew remote
    :param repo_path: Path to directory containing a .git/ directory
    :param name: Name of the remote
    :return:
    """
    repo = git.Repo(repo_path)
    if any(['coldbrew-'+name == remote.name for remote in repo.remotes]):
        repo.delete_remote('coldbrew-'+name)
    else:
        print(f"coldbrew remote {name} does not exist")
        sys.exit(1)


def remote_list(repo_path: Path) -> List[Tuple[str, str]]:
    """
    List all coldbrew remotes
    :param repo_path: Path to directory containing a .git/ directory
    :return:
    """
    repo = git.Repo(repo_path)
    return [(remote.name[9:], remote.url) for remote in repo.remotes if 'coldbrew-' in remote.name]


def is_initialized(repo_path: Path) -> bool:
    """
    Check if coldbrew has been initialized in the given repo path
    :param repo_path: Path to directory containing a .git/ directory
    :return:
    """
    repo = git.Repo(repo_path)
    return 'coldbrew/util' in repo.heads and any(['coldbrew-' in remote.name for remote in repo.remotes])


if __name__ == '__main__':
    raise NotImplementedError("This script is not meant to be run directly")
