#!/usr/bin/env python

import sys
import argparse
import argcomplete
import readline  # Have to import readline to make input() more pleasant for the user (cursor movement etc.)
from pathlib import Path


class ServerCommands:
    @staticmethod
    def init(args):
        """
        Initialize the server: Create empty bare git repository in <args.path>/coldbrew.git/
        :param args: Namespace containing path
        """
        repo_path = args.path / 'coldbrew.git'

        if repo_path.is_dir():
            print(f"coldbrew server has already been initialized in {repo_path}")
            sys.exit(1)

        resp = input(f"Initialize coldbrew server in {args.path}? \n"
                     f"This will create an empty bare git repository in {repo_path}/ [y]/n ")
        if resp == 'y' or resp == '':
            from coldbrew import server
            server.create_bare_repo(repo_path)
            print(f"\N{check mark} Initialized coldbrew server in {repo_path}/")
        else:
            print("Aborted.")
            sys.exit(1)

    @staticmethod
    def run(args):
        """
        Run the server: Start the flask/waitress web server
        :param args: Namespace containing path, address and port
        """
        repo_path = args.path / 'coldbrew.git'

        if not repo_path.is_dir():
            print(f"coldbrew server has not yet been initialized in {args.path}. Run 'coldbrew server init' to initialize here")
            sys.exit(1)

        from coldbrew import server
        server.run(args.path, args.address, args.port)


class RemoteCommands:
    @staticmethod
    def add(args):
        """
        Add a new coldbrew remote with name and url
        :param args: Namespace containing path, name and url of the new remote
        :return:
        """
        from coldbrew import client
        client.remote_add(args.path, args.name, args.url)

    @staticmethod
    def remove(args):
        """
        Remove a coldbrew remote by name
        :param args: Namespace containing path
        :return:
        """
        from coldbrew import client
        client.remote_remove(args.path, args.name)

    @staticmethod
    def list(args):
        """
        Lists all configured coldbrew remotes (git remotes starting with coldbrew-)
        :param args: Namespace containing path
        :return:
        """
        from coldbrew import client
        remotes = client.remote_list(args.path)

        print(f"coldbrew remotes:")
        for remote in remotes:
            print(f" {remote[0]} ({remote[1]})")


class ClientCommands:
    @staticmethod
    def init(args):
        """
        Initialize the client: If in valid working directory, create a new coldbrew remote and initial coldbrew branch.
        :param args: Namespace containing path
        :return:
        """
        from coldbrew import client
        if not client.contains_git_repo(args.path):
            print(f"No git repository was found in {args.path}. Run 'coldbrew init' in your working directory")
            sys.exit(1)

        resp = input(f'Initialize coldbrew in {args.path}? \n'
                     f'This will add a coldbrew ref namespace and a remote with appropriate refspec to your existing repository [y]/n ')
        if resp == 'y' or resp == '':
            client.init(args.path)
            print(f"\N{check mark} Initialized coldbrew in {args.path}")
            print("  Adding an initial coldbrew remote now. You can manage remotes with 'coldbrew remote'")
            url = input(f'  Enter the remote git url (ending with coldbrew.git/): ')
            name = input(f'  Name that remote: ')
            client.remote_add(args.path, name, url)
            print(f"\N{check mark} Added remote {name} ({url})")
        else:
            print("Aborted.")
            sys.exit(1)

    @staticmethod
    def create(args):
        """
        Create a new experiment
        :param args: Namespace containing path, description and group of the new experiment
        """
        from coldbrew import client

        if not client.contains_git_repo(args.path):
            print(f"No git repository was found in {args.path}. Run coldbrew in your working directory")
            sys.exit(1)

        if not client.is_initialized(args.path):
            print(f"coldbrew has not yet been initialized in {args.path}. Run 'coldbrew init' to initialize here")
            sys.exit(1)

        # Ask for description and group if not already supplied
        if args.description is None:
            args.description = input("Description: ")
        if args.group is None:
            args.group = input("Group (leave blank to put into top group): ")

        # Make sure there are no slashes in the group name - subdirectory hierarchies are not supported
        while len(args.group) > 0 and args.group[0] == '/':
            args.group = args.group[1:]
        args.group = args.group.replace('/', '-')

        client.create(args.path, args.description, args.group)
        client.push(args.path)

    @staticmethod
    def iterate(args):
        """
        Iterate on an existing experiment
        :param args: Namespace containing path, parent experiment ID and description of the new iteration
        :return:
        """
        from coldbrew import client

        if not client.contains_git_repo(args.path):
            print(f"No git repository was found in {args.path}. Run coldbrew in your working directory")
            sys.exit(1)

        if not client.is_initialized(args.path):
            print(f"coldbrew has not yet been initialized in {args.path}. Run 'coldbrew init' to initialize here")
            sys.exit(1)

        # Ask for description if not already supplied
        if args.description is None:
            args.description = input("Description: ")

        client.iterate(args.path, args.experiment, args.description)
        client.push(args.path)

    @staticmethod
    def push(args):
        """
        Push to all remotes
        :param args: Namespace containing path
        :return:
        """
        from coldbrew import client

        if not client.contains_git_repo(args.path):
            print(f"No git repository was found in {args.path}. Run coldbrew in your working directory")
            sys.exit(1)

        if not client.is_initialized(args.path):
            print(f"coldbrew has not yet been initialized in {args.path}. Run 'coldbrew init' to initialize here")
            sys.exit(1)

        client.push(args.path)


def main():
    """
    Main entry point for the coldbrew command line utility
    """
    try:
        # Main Parser
        parser = argparse.ArgumentParser(description="冷萃 - Prepares and manages neural experiments without getting in the way ☕️🥶")
        subparsers = parser.add_subparsers(title="commands", description="valid coldbrew subcommands")

        # init
        parser_init = subparsers.add_parser('init', help="Initialize coldbrew for the git repository in the current directory")
        parser_init.set_defaults(func=ClientCommands.init)

        # create
        parser_create = subparsers.add_parser('create', help="Create a new experiment that can later be iterated on")
        parser_create.set_defaults(func=ClientCommands.create)
        parser_create.add_argument('-d', '--description', required=False, type=str,
                                   help="Experiment description, will be message of git commit and displayed in the coldbrew server web interface")
        parser_create.add_argument('-g', '--group', required=False, type=str,
                                   help="Experiment group. Experiments in the same group will be placed in the same subdirectory and grouped together in the coldbrew server web interface")

        # iterate
        parser_iterate = subparsers.add_parser('iterate', help="Iterate on a previously created experiment")
        parser_iterate.set_defaults(func=ClientCommands.iterate)
        parser_iterate.add_argument('-e', '--experiment', required=True, type=str,
                                    help="ID or begin of ID uniquely identifying the experiment to iterate on")
        parser_iterate.add_argument('-d', '--description', required=False, type=str,
                                    help="Description of this iteration, e.g. diff to previous iteration")

        # server
        parser_server = subparsers.add_parser('server', help="coldbrew server management: init and run", description="coldbrew server management: init and run")
        server_subparsers = parser_server.add_subparsers(title="server commands", description="valid coldbrew server management subcommands")
        parser_server_init = server_subparsers.add_parser('init', help="Initialize an empty bare coldbrew git repository in the current directory")
        parser_server_init.set_defaults(func=ServerCommands.init)
        parser_server_run = server_subparsers.add_parser('run', help="Run the coldbrew server. This will expose a web interface at the chosen address and port")
        parser_server_run.set_defaults(func=ServerCommands.run)
        parser_server_run.add_argument('-a', '--address', required=False, default='localhost', type=str,
                                       help="The address the server should listen on")
        parser_server_run.add_argument('-p', '--port', required=False, default=8888, type=int,
                                       help="The port the server should listen on")

        # push
        parser_push = subparsers.add_parser('push', help="Push recent snapshots to all remotes. Useful if a previous push failed for some reason")
        parser_push.set_defaults(func=ClientCommands.push)

        # remote
        parser_remote = subparsers.add_parser('remote', help="Remotes management: add, remove, and list", description="Remotes management: add, remove, and list")
        remote_subparsers = parser_remote.add_subparsers(title="remote commands", description="valid coldbrew remote management subcommands")
        parser_remote_add = remote_subparsers.add_parser('add', help="Add a new remote to push to")
        parser_remote_add.set_defaults(func=RemoteCommands.add)
        parser_remote_add.add_argument('-n', '--name', required=True, type=str,
                                       help="Name of the new remote")
        parser_remote_add.add_argument('-u', '--url', required=True, type=str,
                                       help="Url of the remote, ending with coldbrew.git/")
        parser_remote_remove = remote_subparsers.add_parser('remove', help="Remove an existing remote")
        parser_remote_remove.set_defaults(func=RemoteCommands.remove)
        parser_remote_remove.add_argument('-n', '--name', required=True, default='', type=str,
                                          help="Name of the remote to remove. Get all remotes by calling 'coldbrew remote list'")
        parser_remote_list = remote_subparsers.add_parser('list', help="List existing remotes")
        parser_remote_list.set_defaults(func=RemoteCommands.list)

        argcomplete.autocomplete(parser)
        args = parser.parse_args()
        args.path = Path.cwd()

        if 'func' in args:
            func = args.func
            delattr(args, 'func')
            func(args)
        else:
            if sys.argv[-1] == 'server':
                parser_server.print_help()
            elif sys.argv[-1] == 'remote':
                parser_remote.print_help()
            else:
                parser.print_help()

    except KeyboardInterrupt:
        print("Aborted.")
        sys.exit(1)


if __name__ == '__main__':
    raise NotImplementedError("This script is not meant to be run directly")
