import os
import shutil
from pathlib import Path
import datetime
import json
from typing import List, Tuple
from collections import defaultdict

from flask import Flask
from flask import render_template, request, abort

import git
import waitress
import yaml


app = Flask(__name__)


HASH_ABBREV_LEN = 7
TMP_PATH = Path('/tmp/coldbrew/')
base = Path.cwd()


# Functions called from cli
def create_bare_repo(repo_path: Path) -> None:
    """
    Creates a bare repo at the given repo_path
    :param repo_path: The path to the bare repo (usually ending with something.git/). This path will be created
    :return:
    """
    bare_repo = git.Repo.init(repo_path, bare=True)
    assert bare_repo.bare


def run(base_path: Path, host: str, port: int, development=False) -> None:
    """
    Run the coldbrew flask app either in development mode or behind a waitress server for production, depending on the development flag
    :param base_path: Path containing results subdirectories and coldbrew.git/
    :param host: Host to serve from
    :param port: Port to serve from
    :param development: If true, run flask server directly; otherwise, run behind waitress server
    :return:
    """
    global base
    base = base_path

    if development:
        app.run(host, port)
    else:
        waitress.serve(app, host=host, port=port)


# Functions called by flask
def checkout_commit(target_path: Path, hash: str, parent_hash: str) -> Tuple[str, str]:
    """
    Checkout commit with hash into target_path / group / parent_hash / hash / base
    :param target_path: Base path to checkout to
    :param hash: Hash of the commit to check out
    :param parent_hash: Hash of the parent experiment
    :return:
    """
    try:
        repo = git.Repo(base / 'coldbrew.git')
    except git.InvalidGitRepositoryError:
        raise AssertionError(f"{base} does not contain a valid git repo (Bare repo expected in {base / 'coldbrew.git'}).")

    assert f'coldbrew/{parent_hash}' in repo.heads and hash in [c.hexsha for c in repo.iter_commits(rev=f'coldbrew/{parent_hash}')], "The supplied commit hash is not valid. Check if the last push was successful."

    # Save the current head state
    try:
        current_head = repo.head.reference
    except:
        pass

    target_path_str = str(target_path)
    if not target_path_str[-1] == '/':  # GitPython clone requires a / at the end
        target_path_str = f'{target_path_str}/'

    # Get group from experiment commit and description from iteration commit
    description = yaml.load(repo.commit(hash).message, Loader=yaml.FullLoader)['description']
    group = yaml.load(repo.commit(parent_hash).message, Loader=yaml.FullLoader)['group']

    # Checkout code from the requested commit into a /tmp directory
    target_commit = repo.commit(hash)
    repo.head.reference = target_commit
    if target_path.is_dir():
        shutil.rmtree(target_path, ignore_errors=True)
    repo.clone(path=target_path_str)
    shutil.rmtree(target_path / '.git', ignore_errors=True)

    # Restore the repo head state
    try:
        repo.head.reference = current_head
    except:
        pass

    return description, group


def prepare_directories(hash: str, parent_hash: str) -> dict:
    """
    Prepare code and results directories
    :param hash: Hash of the iteration to check out
    :param parent_hash: Hash of the parent experiment
    :return: Dict containing code and results paths as strings
    """
    # Code goes into /tmp
    code_path = TMP_PATH / hash[:HASH_ABBREV_LEN]

    # Checkout this commit into the defined code directory
    description, group = checkout_commit(code_path, hash, parent_hash)

    # Create result directory structure
    results_path = base / group / parent_hash[:HASH_ABBREV_LEN] / hash[:HASH_ABBREV_LEN] / 'base'
    results_path.mkdir(parents=True)

    # Return the command to be run including all generated directory names
    return {
        'code': str(code_path),
        'results': str(results_path)
    }


def match_experiments_with_fs(experiments: List[dict]) -> List[dict]:
    """
    Match experiments list from retrieve_experiments_from_git() with the file system:
    - Check which ones have already been prepared (if yes but code dir is not present anymore, checkout the code again)
    - Check which configuration subdirectories have already been created
    :param experiments: The list returned from retrieve_experiments_from_git()
    :return: A list with same structure, but with added configurations and slightly modified contents
    """
    insert_id = max([experiment['id'] for experiment in experiments]) + 1

    experiments_output = experiments.copy()

    for experiment in experiments:
        if experiment['type'] == 'experiment':
            continue
        iteration_dir = base / experiment['group'] / experiment['parent_hash'][:HASH_ABBREV_LEN] / experiment['hash']
        experiment['is_prepared'] = (iteration_dir / 'base').is_dir()
        if experiment['is_prepared']:
            experiment['actions'] = ['code', 'addconfig', 'delete']

            # Check if code is still present, and checkout again if it isn't
            if not (TMP_PATH / experiment['hash']).is_dir():
                checkout_commit(TMP_PATH / experiment['hash'], experiment['full_hash'], experiment['parent_hash'])
                print(f"[INFO] Code for experiment {experiment['hash']} was not present, had to checkout from git")

            # Add all configurations for current iteration
            for idx, config_dir in enumerate(sorted(iteration_dir.iterdir(), key=lambda x: os.stat(x).st_mtime, reverse=True)):
                path_code = TMP_PATH / experiment['hash']
                path_results = config_dir
                configuration = {
                    'id': insert_id,
                    'pid': experiment['id'],
                    'hash': f'{experiment["hash"]}-{idx}',
                    'full_hash': f'{experiment["full_hash"]}',
                    'parent_hash': f'{experiment["parent_hash"]}',
                    'description': config_dir.stem,
                    'date': datetime.datetime.fromtimestamp(os.stat(config_dir).st_mtime).strftime('%Y-%m-%d %H:%M:%S'),
                    'actions': ['info', 'results', 'delete'],
                    'type': 'configuration',
                    'group': experiment['group'],
                    'is_prepared': True,
                    'cmd_strings': [],
                    'path_code': str(path_code),
                    'path_results': str(path_results)
                }
                insert_id += 1

                # Templating
                if (Path(path_code) / '.coldbrew').is_file():
                    with open(Path(path_code) / '.coldbrew', 'r') as f:
                        cmd_string_tmpls = [l.replace('"', '&quot;') for l in (line.strip() for line in f) if l]
                    configuration['cmd_strings'].extend([tmpl.format_map(defaultdict(str, code_path=str(path_code).replace(' ', '\\ '), results_path=str(path_results).replace(' ', '\\ '))) for tmpl in cmd_string_tmpls])
                if (base / '.coldbrew').is_file():
                    with open(base / '.coldbrew', 'r') as f:
                        cmd_string_tmpls = [l.replace('"', '&quot;') for l in (line.strip() for line in f) if l]
                    configuration['cmd_strings'].extend([tmpl.format_map(defaultdict(str, code_path=str(path_code).replace(' ', '\\ '), results_path=str(path_results).replace(' ', '\\ '))) for tmpl in cmd_string_tmpls])

                experiments_output.append(configuration)

    return experiments_output


def retrieve_experiments_from_git() -> Tuple[list, list]:
    """
    Retrieves experiments and iterations from the git repoaitory
    :return: One list of experiments and one list of groups
    """
    repo = git.Repo(base / 'coldbrew.git')
    if not any(['coldbrew/' in head.name for head in repo.heads]):
        return [], []  # If repo has not yet been pushed to, return empty list

    groups = []
    experiments = []
    experiment_id = 0

    experiment_heads = [head.name for head in repo.heads if 'coldbrew/' in head.name and not 'coldbrew/util' in head.name]
    for head in experiment_heads:
        experiment_id += 1
        parent_id = experiment_id

        iterations = [commit for commit in repo.iter_commits(rev=head) if 'description:' in commit.message]
        base_commit = iterations[-1]
        base_commit_data = yaml.load(base_commit.message, Loader=yaml.FullLoader)

        if base_commit_data['group'] not in groups:
            groups.append(base_commit_data['group'])

        experiments.append({
            'id': experiment_id,
            'pid': 0,
            'hash': base_commit.hexsha[:HASH_ABBREV_LEN],
            'full_hash': base_commit.hexsha,
            'description': base_commit_data['description'],
            'date': datetime.datetime.fromtimestamp(base_commit.committed_date).strftime('%Y-%m-%d %H:%M:%S'),
            'actions': [],
            'type': 'experiment',
            'is_prepared': True,
            'group': base_commit_data['group']
        })
        for commit_idx, commit in enumerate(iterations):
            commit_data = yaml.load(commit.message, Loader=yaml.FullLoader)
            experiment_id += 1
            experiments.append({
                'id': experiment_id,
                'pid': parent_id,
                'hash': commit.hexsha[:HASH_ABBREV_LEN],
                'full_hash': commit.hexsha,
                'parent_hash': base_commit.hexsha,
                'description': commit_data['description'] if commit_idx < len(iterations) - 1 else 'Base',
                'date': datetime.datetime.fromtimestamp(commit.committed_date).strftime('%Y-%m-%d %H:%M:%S'),
                'actions': ['prepare'],
                'type': 'iteration',
                'group': base_commit_data['group'],
                'is_prepared': False,
                'path_code': str(TMP_PATH / commit.hexsha[:HASH_ABBREV_LEN]),
                'path_results': str(base / base_commit_data['group'] / base_commit.hexsha[:HASH_ABBREV_LEN] / commit.hexsha[:HASH_ABBREV_LEN])
            })

    return experiments, groups


# FLASK ROUTES
@app.route('/code/', defaults={'req_path': '/'})
@app.route('/code/<path:req_path>')
def code(req_path):
    if '/..' in req_path:
        return abort(404)

    abs_path = TMP_PATH / req_path

    hash = req_path.split('/')[0]
    results_dir = None
    for path in base.rglob(''):
        if path.name == hash:
            results_dir = path.absolute()
    if results_dir is None:
        return abort(404)

    if abs_path.is_file():
        with open(abs_path, 'r') as f:
            content = f.read()

        print(f"[INFO] Showing content of file {abs_path}")
        return render_template('file.tmpl',
                               parts=zip(abs_path.relative_to(TMP_PATH).parents, abs_path.relative_to(TMP_PATH).parts),
                               current_path=req_path, content=content, root='code')
    elif abs_path.is_dir():
        files = [(path.relative_to(TMP_PATH), path.relative_to(abs_path), 'dir' if path.is_dir() else 'file') for path in abs_path.iterdir()]
        print(f"[INFO] Showing content of directory {abs_path}")
        return render_template('directory.tmpl',
                               parts=zip(abs_path.relative_to(TMP_PATH).parents, abs_path.relative_to(TMP_PATH).parts),
                               current_path=req_path, files=files, root='code')
    else:
        print(f"[ERROR] Could not show code in '{req_path}'")
        return abort(404)


@app.route('/results/', defaults={'req_path': '/'})
@app.route('/results/<path:req_path>')
def results(req_path):
    if '/..' in req_path:
        return abort(404)

    abs_path = base / req_path

    if abs_path.is_file():
        with open(abs_path, 'r') as f:
            content = f.read()

        print(f"[INFO] Showing content of file {abs_path}")
        return render_template('file.tmpl',
                               parts=zip(abs_path.relative_to(base).parents, abs_path.relative_to(base).parts),
                               current_path=req_path, content=content, root='results')
    elif abs_path.is_dir():
        files = [(path.relative_to(base), path.relative_to(abs_path), 'dir' if path.is_dir() else 'file') for path in abs_path.iterdir()]
        print(f"[INFO] Showing content of directory {abs_path}")
        return render_template('directory.tmpl',
                               parts=zip(abs_path.relative_to(base).parents, abs_path.relative_to(base).parts),
                               current_path=req_path, files=files, root='results')
    else:
        print(f"[ERROR] Could not show results in '{req_path}'")
        return abort(404)


@app.route('/delete', methods=['POST'])
def delete():
    raw = request.get_data().decode("utf-8")
    hash, description, type = raw.split('/')

    if type == 'configuration':
        for path in list(base.rglob('')):
            if path.parent.name == hash and path.name == description:
                results_dir = path.absolute()
                shutil.rmtree(results_dir)
                print(f"[INFO] Deleted '{description}' (directory {results_dir})")
                return ''
    elif type == 'iteration':
        for path in list(base.rglob('')):
            if path.name == hash:
                results_dir = path.absolute()
                shutil.rmtree(results_dir)

                code_dir = TMP_PATH / hash
                if code_dir.is_dir():
                    shutil.rmtree(code_dir)

                print(f"[INFO] Deleted '{description}' (directories {code_dir} and {results_dir})")
                return ''

    print(f"[ERROR] Did not delete '{description}'")
    return ''


@app.route('/prepare', methods=['POST'])
def prepare():
    hash, parent_hash = request.get_data().decode("utf-8").split(',')
    try:
        paths = prepare_directories(hash, parent_hash)
    except AssertionError as e:
        msg = f"[Error] {str(e)}"
        print(msg)
        return msg

    print(f"[INFO] Prepared directories {paths['code']} and {paths['results']}")
    return ''


@app.route('/addconfig', methods=['POST'])
def addconfig():
    hash, parent_hash, group, description = request.get_data().decode("utf-8").split('/')

    # Create result directory structure
    config_path = base / group / parent_hash[:HASH_ABBREV_LEN] / hash / description
    config_path.mkdir(parents=False)

    print(f"[INFO] Prepared directory {config_path}")
    return ''


@app.route('/')
def index():
    experiments, groups = retrieve_experiments_from_git()
    if len(experiments) > 0:
        experiments = match_experiments_with_fs(experiments)

    return render_template('experiments.tmpl',
                           experiments_flat=json.dumps(experiments),
                           experiments=json.dumps({exp['hash']: exp for exp in experiments}),
                           groups=groups,
                           groups_json=json.dumps(groups))


if __name__ == '__main__':
    raise NotImplementedError("This script is not meant to be run directly")
